@@ -48,11 +48,11 @@ use constant	DEFAULT_MAX_SLURP_SIZE => 16 * 1024;	# CSV files <= than this size
 
 =head1 VERSION
 
-Version 0.16
+Version 0.17
 
 =cut
 
-our $VERSION = '0.16';
+our $VERSION = '0.17';
 
 =head1 SYNOPSIS
 
@@ -172,12 +172,13 @@ sub init
 	if(scalar(@_)) {
 		my %args = (ref($_[0]) eq 'HASH') ? %{$_[0]} : @_;
 
-		# $defaults->{'dbname'} ||= $args{'dbname'};
-		# $defaults->{'cache'} ||= $args{'cache'};
-		# $defaults->{'cache_duration'} ||= $args{'cache_duration'};
-		# $defaults->{'directory'} ||= $args{'directory'};
-		# $defaults->{'logger'} ||= $args{'logger'};
-		%defaults = (%defaults, %args)
+		if(($args{'expires_in'} && !$args{'cache_duration'})) {
+			# Compatibility with CHI
+			$args{'cache_duration'} = $args{'expires_in'};
+		}
+
+		%defaults = (%defaults, %args);
+		$defaults{'cache_duration'} ||= '1 hour';
 	}
 
 	return \%defaults
@@ -199,7 +200,11 @@ Place to store results
 
 =item * C<cache_duration>
 
-How long to store results in the cache (default is 1 hour)
+How long to store results in the cache (default is 1 hour).
+
+=item * C<expires_in>
+
+Synonym of C<cache_duration>, for compatibility with C<CHI>.
 
 =item * C<dbname>
 
@@ -210,6 +215,21 @@ The database will be held in a file such as $dbname.csv.
 
 Where the database file is held
 
+=item * C<filename>
+
+Filename containing the data.
+When not given,
+the filename is derived from the tablename
+which in turn comes from the class name.
+
+=item * C<logger>
+
+Takes an optional parameter logger, which is used for warnings and traces.
+Can be an object that understands warn() and trace() messages,
+such as a L<Log::Log4perl> or L<Log::Any> object,
+a reference to code,
+or a filename.
+
 =item * C<max_slurp_size>
 
 CSV/PSV/XML files smaller than this are held in RAM (default is 16K),
@@ -288,7 +308,7 @@ sub new {
 
 =head2	set_logger
 
-Sets class or code reference that will be used for logging.
+Sets the class, code reference, or file that will be used for logging.
 
 =cut
 
@@ -306,6 +326,7 @@ sub set_logger
 
 # Open the database connection based on the specified type (e.g., SQLite, CSV).
 # Read the data into memory or establish a connection to the database file.
+# column_names allows the column names to be overridden on CSV files
 
 sub _open {
 	my $self = shift;
@@ -364,8 +385,11 @@ sub _open {
 				($fin, $slurp_file) = File::pfopen::pfopen($dir, $dbname, 'csv:db', '<');
 			}
 		}
+		if(my $filename = $self->{'filename'} || $defaults{'filename'}) {
+			$slurp_file = File::Spec->catfile($dir, $filename);
+		}
 		if(defined($slurp_file) && (-r $slurp_file)) {
-			close($fin);
+			close($fin) if(defined($fin));
 			$sep_char = $args{'sep_char'};
 
 			$self->_debug(__LINE__, ' of ', __PACKAGE__, ": slurp_file = $slurp_file, sep_char = $sep_char");
@@ -433,8 +457,9 @@ sub _open {
 			# Text::CSV::Slurp->import();
 			# $self->{'data'} = Text::CSV::Slurp->load(file => $slurp_file, %options);
 
+			# Can't slurp when we want to use our own column names as Text::xSV::Slurp has no way to override the names
 			# FIXME: Text::xSV::Slurp can't cope well with quotes in field contents
-			if((-s $slurp_file) <= $self->{'max_slurp_size'}) {
+			if(((-s $slurp_file) <= $self->{'max_slurp_size'}) && !$args{'column_names'}) {
 				if((-s $slurp_file) == 0) {
 					# Empty file
 					$self->{'data'} = {};
@@ -630,6 +655,9 @@ sub selectall_hash
 	my $c;
 	if($c = $self->{cache}) {
 		$key = $query;
+		if(wantarray) {
+			$key .= ' array';
+		}
 		if(defined($query_args[0])) {
 			$key .= ' ' . join(', ', @query_args);
 		}
@@ -752,6 +780,9 @@ sub fetchrow_hashref {
 	}
 	my $key;
 	if(defined($query_args[0])) {
+		if(wantarray) {
+			$key = 'array ';
+		}
 		$key = "fetchrow $query " . join(', ', @query_args);
 	} else {
 		$key = "fetchrow $query";
@@ -911,11 +942,14 @@ sub AUTOLOAD {
 		}
 	} else {
 		if(my $data = $self->{'data'}) {
-			# The data has been read in using Text::xSV::Slurp, and it's a simple query
+			# The data has been read in using Text::xSV::Slurp,
 			#	so no need to do any SQL
+			$self->_debug('AUTOLOAD using slurped data');
 			if($self->{'no_entry'}) {
+				$self->_debug('no_entry is set');
 				my ($key, $value) = %params;
 				if(defined($key)) {
+					$self->_debug("key = $key, value = $value, column = $column");
 					foreach my $row(@{$data}) {
 						if(defined($row->{$key}) && ($row->{$key} eq $value) && (my $rc = $row->{$column})) {
 							if(defined($rc)) {
@@ -926,6 +960,7 @@ sub AUTOLOAD {
 							return $rc
 						}
 					}
+					$self->_debug('not found in slurped data');
 				}
 			} elsif(((scalar keys %params) == 1) && defined(my $key = $params{'entry'})) {
 				# Look up the key
@@ -976,6 +1011,7 @@ sub AUTOLOAD {
 			}
 			return
 		}
+		# Data has not been slurped in
 		if(($self->{'type'} eq 'CSV') && !$self->{no_entry}) {
 			$query = "SELECT DISTINCT $column FROM $table WHERE " . $self->{'id'} . " IS NOT NULL AND entry NOT LIKE '#%'";
 			$done_where = 1;
@@ -1014,15 +1050,42 @@ sub AUTOLOAD {
 	} else {
 		$self->_debug("AUTOLOAD $query");
 	}
+	my $cache;
+	my $key;
+	if($cache = $self->{cache}) {
+		if(wantarray) {
+			$key = 'array ';
+		}
+		if(defined($args[0])) {
+			$key = "fetchrow $query " . join(', ', @args);
+		} else {
+			$key = "fetchrow $query";
+		}
+		if(my $rc = $cache->get($key)) {
+			$self->_debug('cache HIT');
+			return $rc;
+		}
+		$self->_debug('cache MISS');
+	} else {
+		$self->_debug('cache not used');
+	}
 	# my $sth = $self->{$table}->prepare($query) || throw Error::Simple($query);
 	my $sth = $self->{$table}->prepare($query) || croak($query);
 	# $sth->execute(@args) || throw Error::Simple($query);
 	$sth->execute(@args) || croak($query);
 
 	if(wantarray) {
-		return map { $_->[0] } @{$sth->fetchall_arrayref()};
+		my @rc = map { $_->[0] } @{$sth->fetchall_arrayref()};
+		if($cache) {
+			$cache->set($key, \@rc, $self->{'cache_duration'});
+		}
+		return @rc;
 	}
-	return $sth->fetchrow_array();	# Return the first match only
+	my $rc = $sth->fetchrow_array();	# Return the first match only
+	if($cache) {
+		return $cache->set($key, $rc, $self->{'cache_duration'});
+	}
+	return $rc;
 }
 
 sub DESTROY {
@@ -1040,13 +1103,28 @@ sub DESTROY {
 }
 
 # Helper routines for logger()
-sub _log {
+sub _log
+{
 	my ($self, $level, @messages) = @_;
 
 	if(my $logger = $self->{'logger'}) {
 		if(ref($logger) eq 'CODE') {
-			$logger->({ level => $level, message => \@messages });
+			# Code reference
+			$logger->({
+				class => ref($self) // __PACKAGE__,
+				function => (caller(2))[3],
+				line => (caller(1))[2],
+				level => $level,
+				message => \@messages
+			});
+		} elsif(!ref($logger)) {
+			# File
+			if(open(my $fout, '>>', $logger)) {
+				print $fout uc($level), ': ', ref($self) // __PACKAGE__, ' ', (caller(2))[3], (caller(1))[2], join(' ', @messages), "\n";
+				close $fout;
+			}
 		} else {
+			# Object
 			$logger->$level(@messages);
 		}
 	}
